/*
	Copyright (c) 1992
	Digital Equipment Corporation, Maynard, Massachusetts 01754

	The HAL Alpha Assembler is free software; you can redistribute 
	it and/or modify it with the sole restriction that this copyright 
	notice be attached.

	The information in this software is subject to change without notice 
	and should not be construed as a commitment by Digital Equipment
	Corporation.

	The author disclaims all warranties with regard to this software,
	including all implied warranties of merchantability and fitness, in
	no event shall he be liable for any special, indirect or consequential
	damages or any damages whatsoever resulting from loss of use, data, 
	or profits, whether in an action of contract, negligence or other
	tortious action, arising out of or in connection with the use or
	performance of this software.

*/
#include <stdio.h>
#include <stdlib.h>
#include "pp.h"

#define NONE		0
#define LW		4
#define QW		8

#if defined(VMS) || defined(__MSDOS__)
#define ALIGN		NONE
#else
#define ALIGN		LW
#endif

int total_alloc;
int mem_blk;
unsigned int mem_use[MAX_MEMORY_CHUNKS];
char *mem_tab[MAX_MEMORY_CHUNKS];

void init_memmgt()
{
  int i;
  
  for (i=0; i<MAX_MEMORY_CHUNKS; ++i) {
    mem_use[i] = 0;
    mem_tab[i] = NULL;
  }
  
  mem_blk = 0;
  mem_tab[0] = malloc(MEMORY_CHUNK_SIZE);

  total_alloc = 0;
  
  return;
}

void end_memmgt()
{
  int i;
  
  for (i=0; i<mem_blk; ++i)
    free(mem_tab[i]);

  return;
}

#ifdef __STDC__
void *my_alloc(unsigned int size)
#else
void *my_alloc(size)
unsigned int size;
#endif
{
  void *ptr;

  if (ALIGN != NONE) {
    if (size % ALIGN != 0) {
      unsigned int junk = size % ALIGN;
      size += (ALIGN - junk);
    }
  }

  /* record memory allocation */
  total_alloc += size;

  if (size > MEMORY_CHUNK_SIZE) {
    /* if it's that big, just malloc it straight... */
    ptr = malloc(size);
  } else {
    unsigned int junk;

    if ((size + mem_use[mem_blk]) > MEMORY_CHUNK_SIZE) {
      if (++mem_blk >= MAX_MEMORY_CHUNKS) {
	errout(E_ERR, "low_alloc", "overran maximum memory chunks.\n");
	uninit();
	exit(EXIT_FAILURE);
      } else {
	mem_tab[mem_blk] = malloc(MEMORY_CHUNK_SIZE);
	if (mem_tab[mem_blk] == NULL) {
	  errout(E_ERR, "low_alloc", "insufficient virtual memory.\n");
	  uninit();
	  exit(EXIT_FAILURE);
	}
      }
    }
    
    junk = mem_use[mem_blk];
    mem_use[mem_blk] += size;
    
    ptr = mem_tab[mem_blk] + junk;
  }  

  return ptr;
}

#ifdef __STDC__
void *my_calloc(int n, int size)
#else
void *my_calloc(n, size)
int n, size;
#endif
{
  void *temp = my_alloc(n*size);
  
  memset(temp, 0, n*size);
  
  return temp;
}

