/*
	Copyright (c) 1992
	Digital Equipment Corporation, Maynard, Massachusetts 01754

	The HAL Alpha Assembler is free software; you can redistribute 
	it and/or modify it with the sole restriction that this copyright 
	notice be attached.

	The information in this software is subject to change without notice 
	and should not be construed as a commitment by Digital Equipment
	Corporation.

	The author disclaims all warranties with regard to this software,
	including all implied warranties of merchantability and fitness, in
	no event shall he be liable for any special, indirect or consequential
	damages or any damages whatsoever resulting from loss of use, data, 
	or profits, whether in an action of contract, negligence or other
	tortious action, arising out of or in connection with the use or
	performance of this software.

*/

/* V1.4		- added '*' (multiplication) operator */
/* V1.3		- improved memory management */
/* V1.2		- added support of macros, .ascii, .asciz, .iif */
/* V1.1		- added support of commaed expressions in .long, .quad and .address */

#ifdef __STDC__
#define PROTO(text)	text
#else
#define PROTO(text)	()
#endif

#ifndef DEBUG
#define VERSION		"V1.4"
#else
#define VERSION		"V1.4-Debug"
#endif

#define int32 long

#ifdef main
#define DECLARE
#else
#define DECLARE	extern
#endif

#include "avl.h"

#if defined(VMS) || defined(__MSDOS__)
#define EXIT_SUCCESS	1
#define EXIT_FAILURE	0
#define alloca	my_alloc
#else
#define EXIT_SUCCESS	0
#define EXIT_FAILURE	1
#endif

#define MAXINT		((unsigned int32) 0xffffffff)

#ifdef __MSDOS__
#define MAX_MEMORY_CHUNKS      	(64*1024)
#define MEMORY_CHUNK_SIZE	(1024)
#else
#define MAX_MEMORY_CHUNKS      	(512)
#define MEMORY_CHUNK_SIZE	(1024*1024)
#endif

/* these are for avl_mod.c */
#define MAX(a,b)	((a) > (b) ? (a) : (b))
#define NIL(type)	((type *) 0)

#define ALLOC(t)          ((t *)my_alloc(sizeof(t)))
#define ALLOCn(b)         ((char *)my_alloc(b))
#define ALLOC0(n,t)       ((n>0)?(t *)my_calloc(n, sizeof(t)):NULL)
#define ALLOC_STR(l)      ((char *)my_alloc((l)+1)) /* allocate one extra char for NUL-termination */
#define FREE(p)           ;

#define ADD_PUSH	0
#define ADD_APPEND	1

typedef struct quad_ QUAD_T;
struct quad_ {
  unsigned int32 v[2];
};

union token_ {
  QUAD_T q;
  char *s;
};

typedef struct token_list_ TOKEN_LIST;
#define YYSTYPE	TOKEN_LIST *
struct token_list_ {
  int type;
  union token_ tok;
  TOKEN_LIST *next;
};

#define S_LABEL		(-2)
typedef struct symbol_ SYMDEF;
struct symbol_ {
  int incarnation;
};

typedef struct arg_list_ ARGLIST;
struct arg_list_ {
  TOKEN_LIST *def, *act;
  int labelnum;
};

typedef struct macro_stk_ MACRO_STK;
struct macro_stk_ {
  ARGLIST *argl;
  int argn;
  char *name;
  MACRO_STK *next;
};

typedef struct macro_def_ MACRODEF;
struct macro_def_ {
  TOKEN_LIST *name;
  TOKEN_LIST *body;
  TOKEN_LIST *args;
};

typedef struct assign_ ASSIGN;
struct assign_ {
  TOKEN_LIST *name;
  TOKEN_LIST *operand1;
  TOKEN_LIST *operand2;
  TOKEN_LIST *operator;
  ASSIGN *next;
};

#include "lexyy.h"

#define F_FILE		0
#define F_STRING	1
#define F_MACRO		2
typedef struct input_file_ INPUT_FILE;
struct input_file_ {
  int type;
  int repeat;
  FILE *f;
  int lineno;
  char *name;
  YY_BUFFER_STATE buff;
  int yychar;
  YYSTYPE yylval;
  INPUT_FILE *next;
};

#if 0
DECLARE int preprocess_only;
#endif
DECLARE int dot_to_next_inst;
DECLARE int max_macro_size;
DECLARE char *macro_temp;
DECLARE int local_label;
DECLARE int quiet;
DECLARE INPUT_FILE *ifile;
DECLARE INPUT_FILE *cur_ifile;
DECLARE avl_tree *symtree;
DECLARE avl_tree *macrotree;
DECLARE avl_tree *symtab;
DECLARE ASSIGN *defs;
DECLARE MACRO_STK *macstk;
DECLARE TOKEN_LIST *cur_psect, *psect_stk;

DECLARE int line_size;
DECLARE char *line, *lptr;
DECLARE void (*oldsegv)();

#define E_INFO		0
#define E_WARN		1
#define E_ERR		2
#define E_NO_LNUM	0x8000
void errout ();

/* prototypes */
int next_file PROTO((void));
void add_input PROTO((char *name, int type, int repeat, int where));
YYSTYPE cvtradix PROTO((char *text, int radix, int type));
void do_include PROTO((TOKEN_LIST *name_tok));
void do_if PROTO((TOKEN_LIST *cond, TOKEN_LIST *exp, TOKEN_LIST *body, TOKEN_LIST *endbody));
void do_repeat PROTO((TOKEN_LIST *exp, TOKEN_LIST *body));
TOKEN_LIST *eval_const_expr PROTO((TOKEN_LIST *expr));
void output_assignments PROTO((ASSIGN *defs));
char *tok2str PROTO((TOKEN_LIST *temp));
int mul64 PROTO((unsigned int32 *a, unsigned int32 b));
TOKEN_LIST *addq PROTO((TOKEN_LIST *op1, TOKEN_LIST *op2));
TOKEN_LIST *subq PROTO((TOKEN_LIST *op1, TOKEN_LIST *op2));
TOKEN_LIST *mulq PROTO((TOKEN_LIST *op1, TOKEN_LIST *op2));
TOKEN_LIST *notq PROTO((TOKEN_LIST *op1));
TOKEN_LIST *shiftq PROTO((TOKEN_LIST *op1, TOKEN_LIST *op2));
TOKEN_LIST *andq PROTO((TOKEN_LIST *op1, TOKEN_LIST *op2));
TOKEN_LIST *orq PROTO((TOKEN_LIST *op1, TOKEN_LIST *op2));
TOKEN_LIST *xorq PROTO((TOKEN_LIST *op1, TOKEN_LIST *op2));
YYSTYPE make_token PROTO((char *text, int type));
void output_line PROTO((YYSTYPE comment));
void clear_line PROTO((void));
void add_to_line PROTO((char *text));
void *my_alloc PROTO((unsigned int size));
void *my_calloc PROTO((int n, int size));
int is_macro PROTO((char *name));
TOKEN_LIST *copy_expr PROTO((TOKEN_LIST *expr));
void segv_catcher();
void my_free PROTO((void *ptr));
void free_assign PROTO((ASSIGN *defs));
void free_macrodef PROTO((MACRODEF *mtmp));
