/*
	Copyright (c) 1992
	Digital Equipment Corporation, Maynard, Massachusetts 01754

	The HAL Alpha Assembler is free software; you can redistribute 
	it and/or modify it with the sole restriction that this copyright 
	notice be attached.

	The information in this software is subject to change without notice 
	and should not be construed as a commitment by Digital Equipment
	Corporation.

	The author disclaims all warranties with regard to this software,
	including all implied warranties of merchantability and fitness, in
	no event shall he be liable for any special, indirect or consequential
	damages or any damages whatsoever resulting from loss of use, data, 
	or profits, whether in an action of contract, negligence or other
	tortious action, arising out of or in connection with the use or
	performance of this software.

*/
#include <stdio.h>
#include <string.h>
#include <varargs.h>
#include <signal.h>
#include <errno.h>

extern char *sys_errlist[];

#include "pp.h"

#if defined(VMS) || defined(__MSDOS__)
#include "pp_tab.h"
#else
#include "pp.tab.h"
#endif

extern int yychar;
extern YYSTYPE yylval;

#ifdef __STDC__
void free_file(INPUT_FILE *tfil)
#else
void free_file(tfil)
INPUT_FILE *tfil;
#endif
{
  FREE(tfil->name);
  yy_delete_buffer(tfil->buff);
  FREE(tfil);

  return;
}

#ifdef __STDC__
void free_tok(TOKEN_LIST *tok)
#else
void free_tok(tok)
TOKEN_LIST *tok;
#endif
{
  if (tok == NULL)
    return;

  /* free the string assoc with the token (if it was allocated) */
  switch (tok->type) {
  case STRING:
  case SYMBOL:
  case ARG_VALUE:
  case MACRO_CALL:
  case ENDM:
  case MEM_DIRECTIVE:
  case COMMENT:
  case LABEL:
  case IFF:
  case IFT:
  case IFTF:
  case ENDC:
  case ENDR:
    FREE(tok->tok.s);
    break;
  default:
    /* an operator or DOT.  the type is the important part */
    break;
  }

  FREE(tok);

  return;
}

#ifdef __STDC__
void free_tok_list(TOKEN_LIST *tok)
#else
void free_tok_list(tok)
TOKEN_LIST *tok;
#endif
{

  if (tok == NULL)
    return;

  free_tok_list(tok->next);
  free_tok(tok);

  return;
}

int next_file()
{
  INPUT_FILE *temp;

  if (ifile == NULL)
    return 0;

  if ((ifile->type == F_FILE) && (ifile->f == NULL)) {
    ifile->f = fopen(ifile->name, "r");
    if (ifile->f == NULL) {
      errout(E_ERR | E_NO_LNUM, "pp", "%s (file %s)\n", sys_errlist[errno], ifile->name);
      errout(E_ERR | E_NO_LNUM, "pp", "error opening %s for input, aborting.\n", ifile->name);
      return 0;
    }
    if (!quiet)
      errout(E_INFO | E_NO_LNUM, "pp", "opening %s for input.\n", ifile->name);
  }

  if (ifile->buff == NULL)
    ifile->buff = yy_create_buffer(ifile->f, YY_BUF_SIZE);
  
  yy_switch_to_buffer(ifile->buff);
  yychar = ifile->yychar;
  yylval = ifile->yylval;
  
  cur_ifile = ifile;
  ifile = ifile->next;
  
  return 1;
}

#ifdef __STDC__
void add_input(char *name, int type, int repeat, int where)
#else
void add_input(name, type, repeat, where)
char *name;
int type, repeat, where;
#endif
{
  INPUT_FILE **temp, *temp1;

  switch(type) {
  case F_FILE:
    switch (where) {
    case ADD_PUSH:
      temp1 = ALLOC0(1, INPUT_FILE);
      temp1->next = ifile;
      ifile = temp1;
      break;
    case ADD_APPEND:
      for (temp = &ifile; *temp != NULL;)
	temp = &(*temp)->next;
      
      temp1 = *temp = ALLOC0(1, INPUT_FILE);
      break;
    default:
      errout(E_ERR | E_NO_LNUM, "add_input", "invalid value for where argument (%d).\n", where);
      break;
    }
    
    temp1->type = F_FILE;
    temp1->repeat = repeat;
    temp1->name = strcpy(ALLOC_STR(strlen(name)), name);
    temp1->yychar = YYEMPTY;
    break;
  case F_MACRO:
  case F_STRING:
    switch (where) {
    case ADD_PUSH:
      temp1 = ALLOC0(1, INPUT_FILE);
      temp1->next = ifile;
      ifile = temp1;
      break;
    case ADD_APPEND:
      for (temp = &ifile; *temp != NULL;)
	temp = &(*temp)->next;
      
      temp1 = *temp = ALLOC0(1, INPUT_FILE);
      break;
    default:
      errout(E_ERR | E_NO_LNUM, "add_input", "invalid value for where argument (%d).\n", where);
      break;
    }
    
    temp1->type = type;
    temp1->repeat = repeat;
    temp1->name = strcpy(ALLOC_STR(strlen(name)), name);
    temp1->f = (FILE *)temp1->name;
    temp1->yychar = YYEMPTY;
    break;
  default:
    break;
  }
  
  return;
}

#ifdef __STDC__
char *tok2str(TOKEN_LIST *temp)
#else
char *tok2str(temp)
TOKEN_LIST *temp;
#endif
{
  char *retval;

  if (temp == NULL)
    return "";

  switch (temp->type) {
  case NUMBER:
    retval = ALLOC_STR(18);
    sprintf(retval, "^x%08lX%08lX", temp->tok.q.v[1], temp->tok.q.v[0]);
    break;
  case SYMBOL:
    retval = temp->tok.s;
    break;
  case NOT:
    retval = "~";
    break;
  case DOT:
    retval = ".";
    break;
  default:
    retval = ALLOC_STR(1);
    retval[0] = temp->type;
    retval[1] = '\0';
    break;
  }

  return retval;
}

int add64(a,b,c)
int32 *a,*b,*c;
{
 int32 t,t1,ov;
 int32 temp[2];
	
 t = (a[0] & 0xFFFF) + (b[0] & 0xFFFF);
 temp[0] = t & 0xFFFF;

 t1 = ((a[0]>>16) & 0x7FFF) + ((b[0]>>16) & 0x7FFF) + (t >>16);
 t  = ((a[0]>>16) & 0xFFFF) + ((b[0]>>16) & 0xFFFF) + (t >>16);
 ov = (t>>16 & 1) ^ (t1>>15 & 1);
 temp[0] |= t<<16;

 t = (a[1] & 0xFFFF) + (b[1] & 0xFFFF) + (t >>16);
 temp[1] = t & 0xFFFF;

 t1 = ((a[1]>>16) & 0x7FFF) + ((b[1]>>16) & 0x7FFF) + (t >>16);
 t  = ((a[1]>>16) & 0xFFFF) + ((b[1]>>16) & 0xFFFF) + (t >>16);
 ov |= ((t>>16 & 1) ^ (t1>>15 & 1))<<1;
 temp[1] |= t<<16;
    
 c[0] = temp[0];
 c[1] = temp[1];

 return (ov);
}

/* op1 and op2 must be NUMBERs */
#ifdef __STDC__
TOKEN_LIST *addq(TOKEN_LIST *op1, TOKEN_LIST *op2)
#else
TOKEN_LIST *addq(op1, op2)
TOKEN_LIST *op1, *op2;
#endif
{
  TOKEN_LIST *temp = ALLOC0(1, TOKEN_LIST);
  temp->type = NUMBER;

  if (add64(op1->tok.q.v, op2->tok.q.v, temp->tok.q.v)) {
    /* overflow */
    errout(E_WARN, "addq", "overflow in integer add.\n");
  }

  return temp;
}

int sub64(a,b,c)
int32 *a,*b,*c;
{
  int32 t,t1,ov;
  int32 bb[2];
  int32 temp[2];
  
  bb[0] = b[0] ^ 0xFFFFFFFF;
  bb[1] = b[1] ^ 0xFFFFFFFF;
  
  t = (a[0] & 0xFFFF) + (bb[0] & 0xFFFF) +1;
  temp[0] = t & 0xFFFF;
  
  t1 = ((a[0]>>16) & 0x7FFF) + ((bb[0]>>16) & 0x7FFF) + (t >>16);
  t = ((a[0]>>16) & 0xFFFF) + ((bb[0]>>16) & 0xFFFF) + (t >>16);
  ov = (t>>16 & 1) ^ (t1>>15 & 1);
  temp[0] |= t<<16;
  
  t = (a[1] & 0xFFFF) + (bb[1] & 0xFFFF) + (t >>16);
  temp[1] = t & 0xFFFF;
  
  t1 = ((a[1]>>16) & 0x7FFF) + ((bb[1]>>16) & 0x7FFF) + (t >>16);
  t = ((a[1]>>16) & 0xFFFF) + ((bb[1]>>16) & 0xFFFF) + (t >>16);
  ov |= ((t>>16 & 1) ^ (t1>>15 & 1))<<1;
  temp[1] |= t<<16;
  
  c[0] = temp[0];
  c[1] = temp[1];
  
  return (ov);
}

/* op1 and op2 must be NUMBERs */
#ifdef __STDC__
TOKEN_LIST *mulq(TOKEN_LIST *op1, TOKEN_LIST *op2)
#else
TOKEN_LIST *mulq(op1, op2)
TOKEN_LIST *op1, *op2;
#endif
{
  TOKEN_LIST *temp = ALLOC0(1, TOKEN_LIST);
  QUAD_T t;
  temp->type = NUMBER;

  temp->tok.q.v[0] = op1->tok.q.v[0];
  temp->tok.q.v[1] = op1->tok.q.v[1];
  t.v[0] = op1->tok.q.v[0];
  t.v[1] = op1->tok.q.v[1];

  mul64(&temp->tok.q, op2->tok.q.v[0]);
  mul64(&t, op2->tok.q.v[1]);
  temp->tok.q.v[1] =  temp->tok.q.v[1] + t.v[0];

  return temp;
}

/* UNSIGNED 64 bit number times 32 bit number -> 64bit number.
   function returns 0 if overflow, 1 otherwise. */
#ifdef __STDC__
int mul64(unsigned int32 *a, unsigned int32 b)
#else
int mul64(a, b)
unsigned int32 *a, b;
#endif
{
  unsigned int32 d1, d2, d3, d4, i1, i2;
  unsigned int32 temp[2], junk;
  unsigned int32 result[2];
  int ov = 0;

  i1 = b >> 16;
  i2 = b & 0xFFFF;

  d1 = a[1] >> 16;
  d2 = a[1] & 0xFFFF;
  d3 = a[0] >> 16;
  d4 = a[0] & 0xFFFF;

  result[0] = result[1] = 0;

  result[0] = i2 * d4;

  junk = i2 * d3;
  temp[1] = junk >> 16;
  temp[0] = (junk & 0xFFFF) << 16;
  add64(result, temp, result);

  temp[0] = 0;
  temp[1] = i2 * d2;
  add64(result, temp, result);

  junk = i2 * d1;
  ov |= (junk >> 16) != 0;
  temp[0] = 0;
  temp[1] = (junk & 0xFFFF) << 16;
  add64(result, temp, result);

  junk = i1 * d4;
  temp[1] = junk >> 16;
  temp[0] = (junk & 0xFFFF) << 16;
  add64(result, temp, result);

  temp[0] = 0;
  temp[1] = i1 * d3;
  add64(result, temp, result);

  junk = i1 * d2;
  ov |= (junk >> 16) != 0;
  temp[0] = 0;
  temp[1] = (junk & 0xFFFF) << 16;
  add64(result, temp, result);

  junk = i1 * d1;
  ov |= junk != 0;

  a[1] = result[1];
  a[0] = result[0];

  return ov;
}

/* op1 and op2 must be NUMBERs */
#ifdef __STDC__
TOKEN_LIST *subq(TOKEN_LIST *op1, TOKEN_LIST *op2)
#else
TOKEN_LIST *subq(op1, op2)
TOKEN_LIST *op1, *op2;
#endif
{
  TOKEN_LIST *temp = ALLOC0(1, TOKEN_LIST);
  temp->type = NUMBER;

  if (sub64(op1->tok.q.v, op2->tok.q.v, temp->tok.q.v)) {
    /* overflow */
    errout(E_ERR, "subq", "overflow in integer subtract.\n");
  }

  return temp;
}

/* op1 must be a NUMBER */
#ifdef __STDC__
TOKEN_LIST *notq(TOKEN_LIST *op1)
#else
TOKEN_LIST *notq(op1)
TOKEN_LIST *op1;
#endif
{
  TOKEN_LIST *temp = ALLOC0(1, TOKEN_LIST);
  temp->type = NUMBER;

  temp->tok.q.v[0] = ~op1->tok.q.v[0];
  temp->tok.q.v[1] = ~op1->tok.q.v[1];

  return temp;
}

/* op1 and op2 must be NUMBERs */
#ifdef __STDC__
TOKEN_LIST *shiftq(TOKEN_LIST *op1, TOKEN_LIST *op2)
#else
TOKEN_LIST *shiftq(op1, op2)
TOKEN_LIST *op1, *op2;
#endif
{
  TOKEN_LIST *temp = ALLOC0(1, TOKEN_LIST);
  int32 num = (int32)op2->tok.q.v[0];

  temp->type = NUMBER;
  temp->tok.q.v[1] = op1->tok.q.v[1];
  temp->tok.q.v[0] = op1->tok.q.v[0];

  if (num == 0)
    return temp;

  if (num <= 63) {
    /* do the shift, this is REALLY inefficient, but easy */
    int dir = 1, i;

    if (num < 0) {
      dir = 0;
      num = num * -1;
    }

    for (i = 0; i < num; ++i) {
      if (dir) {
	int32 out = temp->tok.q.v[0] >> 31;

	temp->tok.q.v[1] = (temp->tok.q.v[1] << 1) | out;
	temp->tok.q.v[0] = (temp->tok.q.v[0] << 1);
      } else {
	int32 out = temp->tok.q.v[1] << 31;

	temp->tok.q.v[1] = (temp->tok.q.v[1] >> 1);
	temp->tok.q.v[0] = (temp->tok.q.v[0] >> 1) | out;
      }
    }
  }

  return temp;
}

/* op1 and op2 must be NUMBERs */
#ifdef __STDC__
TOKEN_LIST *andq(TOKEN_LIST *op1, TOKEN_LIST *op2)
#else
TOKEN_LIST *andq(op1, op2)
TOKEN_LIST *op1, *op2;
#endif
{
  TOKEN_LIST *temp = ALLOC0(1, TOKEN_LIST);
  temp->type = NUMBER;

  temp->tok.q.v[0] = op1->tok.q.v[0] & op2->tok.q.v[0];
  temp->tok.q.v[1] = op1->tok.q.v[1] & op2->tok.q.v[1];

  return temp;
}

/* op1 and op2 must be NUMBERs */
#ifdef __STDC__
TOKEN_LIST *orq(TOKEN_LIST *op1, TOKEN_LIST *op2)
#else
TOKEN_LIST *orq(op1, op2)
TOKEN_LIST *op1, *op2;
#endif
{
  TOKEN_LIST *temp = ALLOC0(1, TOKEN_LIST);
  temp->type = NUMBER;

  temp->tok.q.v[0] = op1->tok.q.v[0] | op2->tok.q.v[0];
  temp->tok.q.v[1] = op1->tok.q.v[1] | op2->tok.q.v[1];

  return temp;
}

/* op1 and op2 must be NUMBERs */
#ifdef __STDC__
TOKEN_LIST *xorq(TOKEN_LIST *op1, TOKEN_LIST *op2)
#else
TOKEN_LIST *xorq(op1, op2)
TOKEN_LIST *op1, *op2;
#endif
{
  TOKEN_LIST *temp = ALLOC0(1, TOKEN_LIST);
  temp->type = NUMBER;

  temp->tok.q.v[0] = op1->tok.q.v[0] ^ op2->tok.q.v[0];
  temp->tok.q.v[1] = op1->tok.q.v[1] ^ op2->tok.q.v[1];

  return temp;
}

#ifdef __STDC__
int is_macro(char *name)
#else
int is_macro(name)
char *name;
#endif
{
  char *ctmp;

  /* convert to lowercase first */
  for (ctmp = name; *ctmp; ++ctmp)
    *ctmp = tolower(*ctmp);
  
  return avl_lookup(macrotree, name, NULL);
}

void errout(level, rtn, s, va_alist)
int level;
char *rtn, *s;
va_dcl
{
  char temp[2048], lchar;
  va_list the_args;
  
  if (quiet && ((level & ~E_NO_LNUM) == E_INFO))
    return;

  va_start(the_args);

  switch (level & ~E_NO_LNUM) {
  case E_INFO:
    lchar = 'I';
    break;
  case E_WARN:
    lchar = 'W';
    break;
  case E_ERR:
    lchar = 'E';
    break;
  default:
    lchar = '?';
    break;
  }

  sprintf(temp, "%%%%%c-%s, %s", lchar, rtn, s);
  vfprintf(stderr, temp, the_args);

  if (!(level & E_NO_LNUM)) {
    MACRO_STK *mstk = macstk;
    INPUT_FILE *tfil = cur_ifile;
    int is_ok = 1;

    if ((tfil->type != F_FILE) && (tfil->type != F_MACRO)) {
      /* current input file is not a file or macro, search input file stack for the
	 first file or macro, and that is where we are. */
      is_ok = 0;
    }
    
    /* do cur_ifile, then the entire input file stack */
    if ((tfil->type == F_FILE) || (tfil->type == F_MACRO)) {
      if ((tfil->type == F_MACRO) && (mstk == NULL)) {
	fprintf(stderr, "%%E-errout, macro stack doesn't match number of executing macros.\n");
      } else {
	fprintf(stderr, "%%%c-%s, in %s%s on%s line %d.\n", lchar, rtn, (tfil->type == F_FILE)?"file ":"macro ", 
		(tfil->type == F_FILE)?tfil->name:mstk->name, is_ok?"":" or before", tfil->lineno);
	if (tfil->type == F_MACRO)
	  mstk = mstk->next;
      }
    }
    for (tfil = ifile; tfil != NULL; tfil = tfil->next) {
      if ((tfil->type == F_MACRO) && (mstk == NULL)) {
	fprintf(stderr, "%%E-errout, macro stack doesn't match number of executing macros.\n");
	break;
      }
      if ((tfil->type == F_FILE) || (tfil->type == F_MACRO))
	fprintf(stderr, "%%%c-%s, in %s%s on%s line %d.\n", lchar, rtn, (tfil->type == F_FILE)?"file ":"macro ", 
		(tfil->type == F_FILE)?tfil->name:mstk->name, is_ok?"":" or before", tfil->lineno);
      if (tfil->type == F_MACRO) {
	mstk = mstk->next;
      }
    }
    fprintf(stderr, "\n");
  }

  return;
}

void segv_catcher(code)
int code;
{
  signal(SIGSEGV, oldsegv);

  errout(E_ERR, "segv_catcher", "Segmentation fault\n");

  uninit();
  exit(EXIT_FAILURE);
}
